__all__ = [
    "compute_DIC",
    "residuals_plot",
    "summary",
    "trace_plot"
]


from .compute_DIC import compute_DIC
from .residuals_plot import residuals_plot
from .summary import summary
from .trace_plot import trace_plot
